package sun.reflect.annotation;

import java.lang.reflect.Array;
import java.lang.reflect.GenericArrayType;
import java.lang.reflect.Type;
import sun.reflect.generics.tree.TypeSignature;
import sun.reflect.generics.visitor.Reifier;
import sun.reflect.generics.factory.CoreReflectionFactory;
import sun.reflect.generics.scope.ClassScope;
import sun.reflect.generics.parser.SignatureParser;
import java.security.AccessController;
import java.lang.reflect.Proxy;
import java.security.PrivilegedAction;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.lang.annotation.RetentionPolicy;
import java.nio.ByteBuffer;
import java.util.LinkedHashMap;
import java.util.Collections;
import java.nio.BufferUnderflowException;
import java.lang.annotation.AnnotationFormatError;
import java.util.Map;
import sun.reflect.ConstantPool;
import java.lang.annotation.Annotation;

public class AnnotationParser
{
    private static final Annotation[] EMPTY_ANNOTATIONS_ARRAY;
    private static final Annotation[] EMPTY_ANNOTATION_ARRAY;
    
    public AnnotationParser() {
    }
    
    public static Map<Class<? extends Annotation>, Annotation> parseAnnotations(byte[] array, ConstantPool constantPool, Class<?> clazz) {
        if (array == null) {
            return Collections.emptyMap();
        }
        try {
            return parseAnnotations2(array, constantPool, clazz, (Class[])null);
        }
        catch (BufferUnderflowException ex) {
            throw new AnnotationFormatError("Unexpected end of annotations.");
        }
        catch (IllegalArgumentException cause) {
            throw new AnnotationFormatError(cause);
        }
    }
    
    @SafeVarargs
    static Map<Class<? extends Annotation>, Annotation> parseSelectAnnotations(byte[] array, ConstantPool constantPool, Class<?> clazz, Class<? extends Annotation>... array2) {
        if (array == null) {
            return Collections.emptyMap();
        }
        try {
            return parseAnnotations2(array, constantPool, clazz, array2);
        }
        catch (BufferUnderflowException ex) {
            throw new AnnotationFormatError("Unexpected end of annotations.");
        }
        catch (IllegalArgumentException cause) {
            throw new AnnotationFormatError(cause);
        }
    }
    
    private static Map<Class<? extends Annotation>, Annotation> parseAnnotations2(byte[] array, ConstantPool constantPool, Class<?> clazz, Class<? extends Annotation>[] array2) {
        LinkedHashMap<Class, Annotation> linkedHashMap = new LinkedHashMap();
        ByteBuffer wrap = ByteBuffer.wrap(array);
        for (int n = wrap.getShort() & 0xFFFF, i = 0; i < n; ++i) {
            Annotation annotation2 = parseAnnotation2(wrap, constantPool, clazz, false, array2);
            if (annotation2 != null) {
                Class<Object> annotationType = annotation2.annotationType();
                if (AnnotationType.getInstance(annotationType).retention() == RetentionPolicy.RUNTIME && linkedHashMap.put(annotationType, annotation2) != null) {
                    throw new AnnotationFormatError("Duplicate annotation for class: " + annotationType + ": " + annotation2);
                }
            }
        }
        return linkedHashMap;
    }
    
    public static Annotation[][] parseParameterAnnotations(byte[] array, ConstantPool constantPool, Class<?> clazz) {
        try {
            return parseParameterAnnotations2(array, constantPool, clazz);
        }
        catch (BufferUnderflowException ex) {
            throw new AnnotationFormatError("Unexpected end of parameter annotations.");
        }
        catch (IllegalArgumentException cause) {
            throw new AnnotationFormatError(cause);
        }
    }
    
    private static Annotation[][] parseParameterAnnotations2(byte[] array, ConstantPool constantPool, Class<?> clazz) {
        ByteBuffer wrap = ByteBuffer.wrap(array);
        int n = wrap.get() & 0xFF;
        Annotation[][] array2 = new Annotation[n][];
        Object o;
        for (int i = 0; i < n; array2[i] = (Annotation[])o.toArray(AnnotationParser.EMPTY_ANNOTATIONS_ARRAY), ++i) {
            int initialCapacity = wrap.getShort() & 0xFFFF;
            o = new ArrayList(initialCapacity);
            for (int j = 0; j < initialCapacity; ++j) {
                Annotation annotation = parseAnnotation(wrap, constantPool, clazz, false);
                if (annotation != null && AnnotationType.getInstance(annotation.annotationType()).retention() == RetentionPolicy.RUNTIME) {
                    o.add(annotation);
                }
            }
        }
        return array2;
    }
    
    static Annotation parseAnnotation(ByteBuffer byteBuffer, ConstantPool constantPool, Class<?> clazz, boolean b) {
        return parseAnnotation2(byteBuffer, constantPool, clazz, b, (Class[])null);
    }
    
    private static Annotation parseAnnotation2(ByteBuffer byteBuffer, ConstantPool constantPool, Class<?> clazz, boolean b, Class<? extends Annotation>[] array) {
        int n = byteBuffer.getShort() & 0xFFFF;
        String utf8At = "[unknown]";
        Class<Object> clazz2;
        try {
            try {
                utf8At = constantPool.getUTF8At(n);
                clazz2 = parseSig(utf8At, clazz);
            }
            catch (IllegalArgumentException ex) {
                clazz2 = constantPool.getClassAt(n);
            }
        }
        catch (NoClassDefFoundError cause) {
            if (b) {
                throw new TypeNotPresentException(utf8At, cause);
            }
            skipAnnotation(byteBuffer, false);
            return null;
        }
        catch (TypeNotPresentException ex2) {
            if (b) {
                throw ex2;
            }
            skipAnnotation(byteBuffer, false);
            return null;
        }
        if (array != null && !contains(array, clazz2)) {
            skipAnnotation(byteBuffer, false);
            return null;
        }
        AnnotationType instance;
        try {
            instance = AnnotationType.getInstance(clazz2);
        }
        catch (IllegalArgumentException ex3) {
            skipAnnotation(byteBuffer, false);
            return null;
        }
        Map<String, Class<?>> memberTypes = instance.memberTypes();
        LinkedHashMap o = new LinkedHashMap(instance.memberDefaults());
        for (int n2 = byteBuffer.getShort() & 0xFFFF, i = 0; i < n2; ++i) {
            String utf8At2 = constantPool.getUTF8At(byteBuffer.getShort() & 0xFFFF);
            Class<Object> clazz3 = (Class)memberTypes.get(utf8At2);
            if (clazz3 == null) {
                skipMemberValue(byteBuffer);
            }
            else {
                Object memberValue = parseMemberValue(clazz3, byteBuffer, constantPool, clazz);
                if ((memberValue instanceof AnnotationTypeMismatchExceptionProxy)) {
                    ((AnnotationTypeMismatchExceptionProxy)memberValue).setMember((Method)instance.members().get(utf8At2));
                }
                o.put(utf8At2, memberValue);
            }
        }
        return annotationForMap(clazz2, o);
    }
    
    public static Annotation annotationForMap(final Class<? extends Annotation> clazz, final Map<String, Object> map) {
        return (Annotation)AccessController.doPrivileged(new PrivilegedAction<Annotation>() {
            @Override
            public Annotation run() {
                return (Annotation)Proxy.newProxyInstance(clazz.getClassLoader(), new Class[] { clazz }, new AnnotationInvocationHandler(clazz, map));
            }
        });
    }
    
    public static Object parseMemberValue(Class<?> clazz, ByteBuffer byteBuffer, ConstantPool constantPool, Class<?> clazz2) {
        int value = byteBuffer.get();
        Object obj;
        switch (value) {
            case 101: {
                return parseEnumValue(clazz, byteBuffer, constantPool, clazz2);
            }
            case 99: {
                obj = parseClassValue(byteBuffer, constantPool, clazz2);
                break;
            }
            case 64: {
                obj = parseAnnotation(byteBuffer, constantPool, clazz2, true);
                break;
            }
            case 91: {
                return parseArray(clazz, byteBuffer, constantPool, clazz2);
            }
            default: {
                obj = parseConst(value, byteBuffer, constantPool);
                break;
            }
        }
        if (!(obj instanceof ExceptionProxy) && !clazz.isInstance(obj)) {
            obj = new AnnotationTypeMismatchExceptionProxy(obj.getClass() + "[" + obj + "]");
            return obj;
        }
        return obj;
    }
    
    private static Object parseConst(int i, ByteBuffer byteBuffer, ConstantPool constantPool) {
        int n = byteBuffer.getShort() & 0xFFFF;
        switch (i) {
            case 66: {
                return Byte.valueOf((byte)constantPool.getIntAt(n));
            }
            case 67: {
                return Character.valueOf((char)constantPool.getIntAt(n));
            }
            case 68: {
                return Double.valueOf(constantPool.getDoubleAt(n));
            }
            case 70: {
                return Float.valueOf(constantPool.getFloatAt(n));
            }
            case 73: {
                return Integer.valueOf(constantPool.getIntAt(n));
            }
            case 74: {
                return Long.valueOf(constantPool.getLongAt(n));
            }
            case 83: {
                return Short.valueOf((short)constantPool.getIntAt(n));
            }
            case 90: {
                return Boolean.valueOf(constantPool.getIntAt(n) != 0);
            }
            case 115: {
                return constantPool.getUTF8At(n);
            }
            default: {
                throw new AnnotationFormatError("Invalid member-value tag in annotation: " + i);
            }
        }
    }
    
    private static Object parseClassValue(ByteBuffer byteBuffer, ConstantPool constantPool, Class<?> clazz) {
        int n = byteBuffer.getShort() & 0xFFFF;
        try {
            try {
                return parseSig(constantPool.getUTF8At(n), clazz);
            }
            catch (IllegalArgumentException ex) {
                return constantPool.getClassAt(n);
            }
        }
        catch (NoClassDefFoundError noClassDefFoundError) {
            return new TypeNotPresentExceptionProxy("[unknown]", noClassDefFoundError);
        }
        catch (TypeNotPresentException ex2) {
            return new TypeNotPresentExceptionProxy(ex2.typeName(), ex2.getCause());
        }
    }
    
    private static Class<?> parseSig(String s, Class<?> clazz) {
        if (s.equals("V")) {
            return Void.TYPE;
        }
        TypeSignature typeSig = SignatureParser.make().parseTypeSig(s);
        Reifier make = Reifier.make(CoreReflectionFactory.make(clazz, ClassScope.make(clazz)));
        typeSig.accept(make);
        return toClass(make.getResult());
    }
    
    static Class<?> toClass(Type type) {
        if ((type instanceof GenericArrayType)) {
            return Array.newInstance(toClass(((GenericArrayType)type).getGenericComponentType()), 0).getClass();
        }
        return (Class)type;
    }
    
    private static Object parseEnumValue(Class<? extends Enum> enumClass, ByteBuffer byteBuffer, ConstantPool constantPool, Class<?> clazz) {
        String utf8At = constantPool.getUTF8At(byteBuffer.getShort() & 0xFFFF);
        String utf8At2 = constantPool.getUTF8At(byteBuffer.getShort() & 0xFFFF);
        if (!utf8At.endsWith(";")) {
            if (!enumClass.getName().equals(utf8At)) {
                return new AnnotationTypeMismatchExceptionProxy(utf8At + "." + utf8At2);
            }
        }
        else if (enumClass != parseSig(utf8At, clazz)) {
            return new AnnotationTypeMismatchExceptionProxy(utf8At + "." + utf8At2);
        }
        try {
            return Enum.valueOf(enumClass, utf8At2);
        }
        catch (IllegalArgumentException ex) {
            return new EnumConstantNotPresentExceptionProxy(enumClass, utf8At2);
        }
    }
    
    private static Object parseArray(Class<?> clazz, ByteBuffer byteBuffer, ConstantPool constantPool, Class<?> clazz2) {
        int n = byteBuffer.getShort() & 0xFFFF;
        Class<Object> componentType = clazz.getComponentType();
        if (componentType == Byte.TYPE) {
            return parseByteArray(n, byteBuffer, constantPool);
        }
        if (componentType == Character.TYPE) {
            return parseCharArray(n, byteBuffer, constantPool);
        }
        if (componentType == Double.TYPE) {
            return parseDoubleArray(n, byteBuffer, constantPool);
        }
        if (componentType == Float.TYPE) {
            return parseFloatArray(n, byteBuffer, constantPool);
        }
        if (componentType == Integer.TYPE) {
            return parseIntArray(n, byteBuffer, constantPool);
        }
        if (componentType == Long.TYPE) {
            return parseLongArray(n, byteBuffer, constantPool);
        }
        if (componentType == Short.TYPE) {
            return parseShortArray(n, byteBuffer, constantPool);
        }
        if (componentType == Boolean.TYPE) {
            return parseBooleanArray(n, byteBuffer, constantPool);
        }
        if (componentType == String.class) {
            return parseStringArray(n, byteBuffer, constantPool);
        }
        if (componentType == Class.class) {
            return parseClassArray(n, byteBuffer, constantPool, clazz2);
        }
        if (componentType.isEnum()) {
            return parseEnumArray(n, componentType, byteBuffer, constantPool, clazz2);
        }
        assert !(!componentType.isAnnotation());
        return parseAnnotationArray(n, componentType, byteBuffer, constantPool, clazz2);
    }
    
    private static Object parseByteArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool) {
        byte[] array = new byte[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 66) {
                array[i] = (byte)constantPool.getIntAt(byteBuffer.getShort() & 0xFFFF);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseCharArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool) {
        char[] array = new char[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 67) {
                array[i] = (char)constantPool.getIntAt(byteBuffer.getShort() & 0xFFFF);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseDoubleArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool) {
        double[] array = new double[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 68) {
                array[i] = constantPool.getDoubleAt(byteBuffer.getShort() & 0xFFFF);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseFloatArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool) {
        float[] array = new float[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 70) {
                array[i] = constantPool.getFloatAt(byteBuffer.getShort() & 0xFFFF);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseIntArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool) {
        int[] array = new int[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 73) {
                array[i] = constantPool.getIntAt(byteBuffer.getShort() & 0xFFFF);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseLongArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool) {
        long[] array = new long[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 74) {
                array[i] = constantPool.getLongAt(byteBuffer.getShort() & 0xFFFF);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseShortArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool) {
        short[] array = new short[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 83) {
                array[i] = (short)constantPool.getIntAt(byteBuffer.getShort() & 0xFFFF);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseBooleanArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool) {
        boolean[] array = new boolean[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 90) {
                array[i] = (constantPool.getIntAt(byteBuffer.getShort() & 0xFFFF) != 0);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseStringArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool) {
        String[] array = new String[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 115) {
                array[i] = constantPool.getUTF8At(byteBuffer.getShort() & 0xFFFF);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseClassArray(int n, ByteBuffer byteBuffer, ConstantPool constantPool, Class<?> clazz) {
        Class[] array = new Class[n];
        boolean n2 = false;
        int value = 0;
        for (int i = 0; i < n; ++i) {
            value = byteBuffer.get();
            if (value == 99) {
                array[i] = parseClassValue(byteBuffer, constantPool, clazz);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n2 = true;
            }
        }
        return n2 ? exceptionProxy(value) : array;
    }
    
    private static Object parseEnumArray(int length, Class<? extends Enum<?>> componentType, ByteBuffer byteBuffer, ConstantPool constantPool, Class<?> clazz) {
        Object[] array = (Object[])(Object[])Array.newInstance(componentType, length);
        boolean n = false;
        int value = 0;
        for (int i = 0; i < length; ++i) {
            value = byteBuffer.get();
            if (value == 101) {
                array[i] = parseEnumValue(componentType, byteBuffer, constantPool, clazz);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n = true;
            }
        }
        return n ? exceptionProxy(value) : array;
    }
    
    private static Object parseAnnotationArray(int length, Class<? extends Annotation> componentType, ByteBuffer byteBuffer, ConstantPool constantPool, Class<?> clazz) {
        Object[] array = (Object[])(Object[])Array.newInstance(componentType, length);
        boolean n = false;
        int value = 0;
        for (int i = 0; i < length; ++i) {
            value = byteBuffer.get();
            if (value == 64) {
                array[i] = parseAnnotation(byteBuffer, constantPool, clazz, true);
            }
            else {
                skipMemberValue(value, byteBuffer);
                n = true;
            }
        }
        return n ? exceptionProxy(value) : array;
    }
    
    private static ExceptionProxy exceptionProxy(int i) {
        return new AnnotationTypeMismatchExceptionProxy("Array with component tag: " + i);
    }
    
    private static void skipAnnotation(ByteBuffer byteBuffer, boolean b) {
        if (b) {
            byteBuffer.getShort();
        }
        for (int n = byteBuffer.getShort() & 0xFFFF, i = 0; i < n; ++i) {
            byteBuffer.getShort();
            skipMemberValue(byteBuffer);
        }
    }
    
    private static void skipMemberValue(ByteBuffer byteBuffer) {
        skipMemberValue((int)byteBuffer.get(), byteBuffer);
    }
    
    private static void skipMemberValue(int n, ByteBuffer byteBuffer) {
        switch (n) {
            case 101: {
                byteBuffer.getInt();
                break;
            }
            case 64: {
                skipAnnotation(byteBuffer, true);
                break;
            }
            case 91: {
                skipArray(byteBuffer);
                break;
            }
            default: {
                byteBuffer.getShort();
                break;
            }
        }
    }
    
    private static void skipArray(ByteBuffer byteBuffer) {
        for (int n = byteBuffer.getShort() & 0xFFFF, i = 0; i < n; ++i) {
            skipMemberValue(byteBuffer);
        }
    }
    
    private static boolean contains(Object[] array, Object o) {
        int length = array.length;
        int i = 0;
        while (i < length) {
            if (array[i] == o) {
                return true;
            }
            ++i;
        }
        return false;
    }
    
    public static Annotation[] toArray(Map<Class<? extends Annotation>, Annotation> map) {
        return (Annotation[])map.values().toArray(AnnotationParser.EMPTY_ANNOTATION_ARRAY);
    }
    
    static Annotation[] getEmptyAnnotationArray() {
        return AnnotationParser.EMPTY_ANNOTATION_ARRAY;
    }
    
    static {
        EMPTY_ANNOTATIONS_ARRAY = new Annotation[0];
        EMPTY_ANNOTATION_ARRAY = new Annotation[0];
    }
}
